/*
 *	emu.c
 *
 *	Famicom/EMU
 *
 *	* Thu Feb 03 00:00:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

static int
famicom_init(const char* fname)
{
	FILE* fp = NULL;
	//
	int retval;
	INESHEADER hdr;

	/* CVRAMNA܂B */
	memset(m6502mem, 0, sizeof m6502mem);
	memset(ppu_vram, 0, sizeof ppu_vram);

	/* ROMt@CJ܂B */
	fp = fopen(fname, "rb");
	if(!fp) goto L_ERR;

	/* wb_ǂݍ݂܂B */
	retval = fread(&hdr, sizeof hdr, 1, fp);
	if(retval != 1) goto L_ERR;
	if(memcmp(hdr.nes, "NES", 3) != 0) goto L_ERR;
	if(hdr._1a != '\x1a') goto L_ERR;

	/* PRG-ROMǂݍ݂܂B */
	switch(hdr.prgrom_page_count) {
	case 1:
		retval = fread(&m6502mem[0x8000], 0x4000, 1, fp);
		if(retval != 1) goto L_ERR;
		memcpy(&m6502mem[0xc000], &m6502mem[0x8000], 0x4000);
		break;
	case 2:
		retval = fread(&m6502mem[0x8000], 0x4000, 2, fp);
		if(retval != 2) goto L_ERR;
		break;
	default:
		goto L_ERR;
	}

	/* CHR-ROMǂݍ݂܂B */
	switch(hdr.chrrom_page_count) {
	case 1:
		retval = fread(&ppu_vram[0x0000], 0x2000, 1, fp);
		if(retval != 1) goto L_ERR;
		break;
	default:
		goto L_ERR;
	}

	/* ROMt@C܂B */
	fclose(fp);
	fp = NULL;

	/* CPUPPUZbg܂B */
	m6502cpu_reset(famicom_read, famicom_write);
	ppu_reset(hdr.rom_control1 & 1);

	return 0;

L_ERR:
	if(fp) {
		fclose(fp);
	}
	return -1;
}

/****************************************************************************
 *	
 ****************************************************************************/

int
famicom_run(const char* fname)
{
	int retval;
	int cnt;
	int cycle;
	unsigned char* draw;

	/*****  *****/

	retval = famicom_init(fname);
	if(retval != 0) {
		return retval;
	}

	/***** @s *****/

	cnt = 0;
	draw = NULL;
	for(;;) {
		cnt += FPS;
		if(cnt >= 60) {
			cnt -= 60;
			draw = vbuff;
		}

		ppu.flags &= ~PPU_FLAGS_INTERRUPT;
		do {
			cycle = ppu_update(draw);
			m6502cpu_run(cycle / 2);
		} while(!(ppu.flags & PPU_FLAGS_INTERRUPT));
		m6502cpu_nmi();

		if(draw) {
			draw = NULL;
			ppu_reduce(vbuff, vbuff);

			schedule();
			if((joy & PAD_START) && (joy & PAD_SELECT)) {
				break;
			}

			/*  0:A    1:B    2:SEL  3:ST   4:UP  5:DN  6:LF  7:RI
			 *  8:?    9:?   10:?   11:?   12:?  13:?  14:?  15:?
			 * 16:Sig 17:Sig 18:Sig 19:Sig 20:0  21:0  22:0  23:0
			 */
			input = 1 << 16;
			if(joy & PAD_A     ) input |= 1 << 0;
			if(joy & PAD_B     ) input |= 1 << 1;
			if(joy & PAD_SELECT) input |= 1 << 2;
			if(joy & PAD_START ) input |= 1 << 3;
			if(joy & PAD_UP    ) input |= 1 << 4;
			if(joy & PAD_DN    ) input |= 1 << 5;
			if(joy & PAD_LF    ) input |= 1 << 6;
			if(joy & PAD_RI    ) input |= 1 << 7;
		}
	}

	return 0;
}

